#if 0
TBL_MAJOR_VERSION = 2
TBL_MINOR_VERSION = 10
PACKAGE_VERSION     = $(TBL_MAJOR_VERSION).$(TBL_MINOR_VERSION)
#endif
#define TBL_MAJOR_VERSION 2
#define TBL_MINOR_VERSION 10
#define PACKAGE_VERSION "2.10"
